import numpy as np
from scipy.constants import speed_of_light

from .. import PyAMLException
from ..common import abstract
from ..common.element import Element, ElementConfigModel, __pyaml_repr__
from ..configuration import Factory
from ..control.deviceaccess import DeviceAccess
from .function_mapping import function_map
from .magnet import Magnet, MagnetConfigModel
from .model import MagnetModel

# Define the main class name for this module
PYAMLCLASS = "SerializedMagnets"


class ConfigModel(ElementConfigModel):
    function: str
    """List of magnets"""
    elements: list[str] | str
    """List of magnets"""
    model: MagnetModel | None = None
    """Object in charge of converting magnet strengths to currents"""


class ReadWriteSerializedStrengths(abstract.ReadWriteFloatScalar):
    def __init__(self, cfg: ConfigModel, elements: list[abstract.ReadWriteFloatScalar]):
        self.elements = elements
        self._cfg = cfg

    def get(self) -> float:
        return self.elements[0].get()

    def set(self, value: float):
        self.elements[0].set(value)

    def set_and_wait(self, value: float):
        raise NotImplementedError("Not implemented yet.")

    def unit(self) -> str:
        return self._cfg.model.get_strength_units()[0]

    def get_model(self) -> MagnetModel:
        return self._cfg.model

    def get_elements(self):
        return self.elements

    def set_magnet_rigidity(self, brho: np.double):
        [element.set_magnet_rigidity(brho) for element in self.elements]


class ReadWriteSerializedHardwares(ReadWriteSerializedStrengths):
    def __init__(self, cfg: ConfigModel, elements: list[abstract.ReadWriteFloatScalar]):
        super().__init__(cfg, elements)

    def unit(self) -> str:
        return self._cfg.model.get_hardware_units()[0]

    def set_magnet_rigidity(self, brho: np.double):
        [element.set_magnet_rigidity(brho) for element in self.elements]


class SerializedMagnets(Element):
    """
    Class managing serialized magnets: a set of magnet with the same set point.
    The set point is usually managed by only one power supply but it can be covered by several ones.
    If several power supplies


    Parameters
    ----------
    cfg : ConfigModel
        Configuration object TODO: to describe

    Raises
    ------
    pyaml.PyAMLException
        In case of wrong initialization
    """

    def __init__(self, cfg: ConfigModel, peer=None):
        super().__init__(cfg.name)
        self._cfg = cfg
        self.model = cfg.model
        self.polynom = None
        self.__strengths = None
        self.__hardwares = None
        self.__virtuals: list[Magnet] = []
        self.__elements = (
            cfg.elements if isinstance(cfg.elements, list) else [cfg.elements]
        )
        self.model.set_number_of_magnets(len(self.__elements))
        if peer is None:
            # Configuration part
            self.polynom = function_map[self._cfg.function].polynom
            if self._cfg.function not in function_map:
                raise PyAMLException(
                    self._cfg.function + " not implemented for serialized magnet"
                )
            for element in self.__elements:
                # Check mapping validity
                # Create the virtual magnet for the corresponding magnet
                vm = self.__create_virtual_magnet(element)
                self.__virtuals.append(vm)
                # Register the virtual element in the factory to have a coherent factory and improve error reporting
                Factory.register_element(vm)
        else:
            # Attach
            self._peer = peer

    def __create_virtual_magnet(self, name: str) -> Magnet:
        args = {"name": name, "model": self.model}
        virtual: Magnet = function_map[self._cfg.function](MagnetConfigModel(**args))
        virtual.set_model_name(self.get_name())
        return virtual

    def get_nb_magnets(self) -> int:
        return len(self.__elements)

    def get_magnets(self) -> list[Magnet]:
        return self.__virtuals

    def attach(
        self,
        peer,
        strengths: list[abstract.ReadWriteFloatScalar],
        hardwares: list[abstract.ReadWriteFloatScalar],
    ) -> list[Magnet]:
        l = []
        n_ser_mag = SerializedMagnets(self._cfg, peer)
        n_ser_mag.__strengths = ReadWriteSerializedStrengths(self._cfg, strengths)
        n_ser_mag.__hardwares = ReadWriteSerializedHardwares(self._cfg, hardwares)
        l.append(n_ser_mag)
        # Construct a single function magnet for each multipole of this combined function magnet
        for idx, magnet in enumerate(self.__elements):
            strength = strengths[idx]
            hardware = hardwares[idx] if self.model.has_hardware() else None
            l.append(self.__virtuals[idx].attach(peer, strength, hardware))
        return l

    @property
    def strengths(self) -> abstract.ReadWriteFloatScalar:
        """
        Gives access to the strengths of this combined function magnet in physics unit
        """
        self.check_peer()
        if self.__strengths is None:
            raise PyAMLException(
                f"{str(self)} has no model that supports physics units"
            )
        return self.__strengths

    @property
    def hardwares(self) -> abstract.ReadWriteFloatScalar:
        """
        Gives access to the strengths of this combined function magnet in hardware unit when possible
        """
        self.check_peer()
        if self.__hardwares is None:
            raise PyAMLException(
                f"{str(self)} has no model that supports hardware units"
            )
        return self.__hardwares

    def set_energy(self, energy: float):
        brho = energy / speed_of_light
        if self.model is not None:
            self.model.set_magnet_rigidity(brho)
        if self.__hardwares is not None:
            self.__hardwares.set_magnet_rigidity(brho)
        if self.__strengths is not None:
            self.__strengths.set_magnet_rigidity(brho)

    def __repr__(self):
        return __pyaml_repr__(self)

    def get_devices(self) -> list[DeviceAccess]:
        if isinstance(self.model.powerconverter, list):
            return self.model.powerconverter
        else:
            return [self._cfg.powerconverter]



import numpy as np
from pydantic import BaseModel, ConfigDict

from ..common.element import __pyaml_repr__
from ..common.exception import PyAMLException
from ..configuration.curve import Curve
from ..configuration.inline_curve import ConfigModel as InlineCurveModel
from ..configuration.inline_curve import InlineCurve
from ..configuration.matrix import Matrix
from ..control.deviceaccess import DeviceAccess
from .linear_model import ConfigModel as LinearConfigModel
from .linear_model import LinearMagnetModel
from .model import MagnetModel

# Define the main class name for this module
PYAMLCLASS = "LinearSerializedMagnetModel"


class ConfigModel(BaseModel):
    """
    Configuration model for linear serialized magnet model

    Parameters
    ----------
    curves : Curve or list[Curve]
        Excitation curves, 1 curve for all or 1 curve per magnet
    calibration_factors : float or list[float], optional
        Correction factor applied to curves, 1 factor for all or 1 factor per magnet.
        Default: ones
    calibration_offsets : float or list[float], optional
        Correction offset applied to curves, 1 offset for all or 1 offset per magnet.
        Default: zeros
    crosstalk : float or list[float], optional
        Crosstalk factors. Default: 1.0
    powerconverter : DeviceAccess
        The hardware can be a single power supply or a list of power supplies.
        If a list is provided, the same value will be affected to all of them
    unit : str
        Strength unit: rad, m-1, m-2
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    curves: Curve | list[Curve]
    calibration_factors: float | list[float] = None
    calibration_offsets: float | list[float] = None
    crosstalk: float | list[float] = 1.0
    powerconverter: DeviceAccess
    unit: str


def _get_length(elem) -> int:
    if elem is None:
        return 0
    if isinstance(elem, list):
        return len(elem)
    else:
        return 1


def _get_max_length(*args, **kwargs) -> int:
    max_args = max([_get_length(elem) for elem in args]) if args else 0
    max_kwargs = max([_get_length(elem) for elem in kwargs.values()]) if kwargs else 0
    return max(max_args, max_kwargs)


def _to_list_of_length(elem, length: int) -> list:
    if isinstance(elem, list):
        return elem
    else:
        return [elem] * length


def _check_len(obj, name, expected_length):
    length = len(obj)
    if length != expected_length:
        raise PyAMLException(
            f"{name} does not have the expected "
            f"number of items ({expected_length} items expected but got {length})"
        )


class LinearSerializedMagnetModel(MagnetModel):
    """
    Class providing a simple linear model for combined function magnets. A matrix
    can handle separation of multipoles. A pseudo current is a linear combination
    of power supply currents associated to a single function.
    """

    def __init__(self, cfg: ConfigModel):
        self._cfg = cfg
        self.__brho = np.nan

        # Check config
        self.__nbMagnets: int = _get_max_length(
            cfg.curves, cfg.calibration_factors, cfg.calibration_offsets, cfg.crosstalk
        )
        self.__calibration_factors = np.ones(self.__nbMagnets)
        self.__calibration_offsets = np.ones(self.__nbMagnets)
        self.__crosstalk = np.ones(self.__nbMagnets)
        self.__curves = _to_list_of_length(self._cfg.curves, self.__nbMagnets)
        self.__sub_models: list[LinearMagnetModel] = []

    def __initialize(self):
        if self._cfg.calibration_factors is None:
            self.__calibration_factors = np.ones(self.__nbMagnets)
        else:
            self.__calibration_factors = _to_list_of_length(
                self._cfg.calibration_factors, self.__nbMagnets
            )

        if self._cfg.calibration_offsets is None:
            self.__calibration_offsets = np.zeros(self.__nbMagnets)
        else:
            self.__calibration_offsets = _to_list_of_length(
                self._cfg.calibration_offsets, self.__nbMagnets
            )

        if self._cfg.crosstalk is None:
            self.__crosstalk = np.zeros(self.__nbMagnets)
        else:
            self.__crosstalk = _to_list_of_length(self._cfg.crosstalk, self.__nbMagnets)
        self.__curves = _to_list_of_length(self._cfg.curves, self.__nbMagnets)
        if isinstance(self._cfg.curves, list):
            self.__curves = self._cfg.curves
        else:
            self.__curves: list[Curve] = []
            for _ in range(self.__nbMagnets):
                curve = InlineCurve(InlineCurveModel(mat=self._cfg.curves.get_curve()))
                self.__curves.append(curve)

        _check_len(self.__calibration_factors, "calibration_factors", self.__nbMagnets)
        _check_len(self.__calibration_offsets, "calibration_offsets", self.__nbMagnets)
        _check_len(self.__crosstalk, "crosstalk", self.__nbMagnets)
        _check_len(self.__curves, "curves", self.__nbMagnets)

        self.__sub_models: list[LinearMagnetModel] = []
        for magnet_idx in range(self.__nbMagnets):
            sub_model = LinearConfigModel(
                curve=self.__curves[magnet_idx],
                calibration_factor=self.__calibration_factors[magnet_idx],
                calibration_offset=self.__calibration_offsets[magnet_idx],
                crosstalk=self.__crosstalk[magnet_idx],
                powerconverter=self._cfg.powerconverter,
                unit=self._cfg.unit,
            )
            self.__sub_models.append(LinearMagnetModel(sub_model))

    def set_number_of_magnets(self, nb_magnets: int):
        self.__nbMagnets = nb_magnets
        self.__initialize()

    def get_sub_model(self, index: int) -> LinearMagnetModel:
        return self.__sub_models[index]

    def compute_hardware_values(self, strengths: np.array) -> np.array:
        return np.array(
            [
                model.compute_hardware_values([strength])
                for strength, model in zip(strengths, self.__sub_models)
            ]
        )

    def compute_strengths(self, currents: np.array) -> np.array:
        return np.array(
            [
                model.compute_strengths([current])
                for current, model in zip(currents, self.__sub_models)
            ]
        )

    def get_strength_units(self) -> list[str]:
        return self._cfg.units

    def get_hardware_units(self) -> list[str]:
        return [p.unit() for p in self._cfg.__sub_models]

    def get_devices(self) -> list[DeviceAccess]:
        return self._cfg.powerconverters

    def set_magnet_rigidity(self, brho: np.double):
        self.__brho = brho
        [model.set_magnet_rigidity(brho) for model in self.__sub_models]

    def get_magnet_rigidity(self) -> np.double:
        return self.__brho

    def __repr__(self):
        return __pyaml_repr__(self)



"""
Accelerator class
"""

import os

from pydantic import BaseModel, ConfigDict

from .arrays.array import ArrayConfig
from .common.element import Element
from .common.exception import PyAMLConfigException
from .configuration.factory import Factory
from .configuration.fileloader import load, set_root_folder
from .control.controlsystem import ControlSystem
from .lattice.simulator import Simulator

# Define the main class name for this module
PYAMLCLASS = "Accelerator"


class ConfigModel(BaseModel):
    """
    Configuration model for Accelerator

    Parameters
    ----------
    facility : str
        Facility name
    machine : str
        Accelerator name
    energy : float
        Accelerator nominal energy. For ramped machine,
        this value can be dynamically set
    controls : list[ControlSystem], optional
        List of control system used. An accelerator
        can access several control systems
    simulators : list[Simulator], optional
        Simulator list
    data_folder : str
        Data folder
    arrays : list[ArrayConfig], optional
        Element family
    devices : list[Element]
        Element list
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    facility: str
    machine: str
    energy: float
    controls: list[ControlSystem] = None
    simulators: list[Simulator] = None
    data_folder: str
    arrays: list[ArrayConfig] = None
    devices: list[Element]


class Accelerator(object):
    """PyAML top level class"""

    def __init__(self, cfg: ConfigModel):
        self._cfg = cfg
        __design = None
        __live = None

        if cfg.controls is not None:
            for c in cfg.controls:
                if c.name() == "live":
                    self.__live = c
                else:
                    # Add as dynacmic attribute
                    setattr(self, c.name(), c)
                c.fill_device(cfg.devices)

        if cfg.simulators is not None:
            for s in cfg.simulators:
                if s.name() == "design":
                    self.__design = s
                else:
                    # Add as dynacmic attribute
                    setattr(self, s.name(), s)
                s.fill_device(cfg.devices)

        if cfg.arrays is not None:
            for a in cfg.arrays:
                if cfg.simulators is not None:
                    for s in cfg.simulators:
                        a.fill_array(s)
                if cfg.controls is not None:
                    for c in cfg.controls:
                        a.fill_array(c)

        if cfg.energy is not None:
            self.set_energy(cfg.energy)

        self.post_init()

    def set_energy(self, E: float):
        if self._cfg.simulators is not None:
            for s in self._cfg.simulators:
                s.set_energy(E)
        if self._cfg.controls is not None:
            for c in self._cfg.controls:
                c.set_energy(E)

    def post_init(self):
        """
        Method triggered after all initialisations are done
        """
        if self._cfg.simulators is not None:
            for s in self._cfg.simulators:
                s.post_init()
        if self._cfg.controls is not None:
            for c in self._cfg.controls:
                c.post_init()

    @property
    def live(self) -> ControlSystem:
        return self.__live

    @property
    def design(self) -> Simulator:
        return self.__design

    @staticmethod
    def from_dict(config_dict: dict, ignore_external=False) -> "Accelerator":
        """
        Construct an accelerator from a dictionary.
        Parameters
        ----------
        config_dict : str
            Dictionnary conatining accelerator config
        ignore_external: bool
            Ignore external modules and return None for object that
            cannot be created. pydantic schema that support that an
            object is not created should handle None fields.

        """
        if ignore_external:
            # control systems are external, so remove controls field
            config_dict.pop("controls", None)
        return Factory.depth_first_build(config_dict, ignore_external)

    @staticmethod
    def load(
        filename: str, use_fast_loader: bool = False, ignore_external=False
    ) -> "Accelerator":
        """
        Load an accelerator from a config file.

        Parameters
        ----------
        filename : str
            Configuration file name, yaml or json.
        use_fast_loader : bool
            Use fast yaml loader. When specified,
            no line number are reported in case of error,
            only the element name that triggered the error
            will be reported in the exception)
        ignore_external : bool
            Ignore external modules and return None for object that
            cannot be created. pydantic schema that support that an
            object is not created should handle None fields.
        """
        # Asume that all files are referenced from
        # folder where main AML file is stored
        if not os.path.exists(filename):
            raise PyAMLConfigException(f"{filename} file not found")
        rootfolder = os.path.abspath(os.path.dirname(filename))
        set_root_folder(rootfolder)
        config_dict = load(os.path.basename(filename), None, use_fast_loader)
        return Accelerator.from_dict(config_dict)



import os
import time

import numpy as np
from tango.pyaml.attribute import Attribute
from tango.pyaml.attribute import ConfigModel as AttributeConfig
from tango.pyaml.attribute_read_only import (
    AttributeReadOnly,
)
from tango.pyaml.attribute_read_only import (
    ConfigModel as AttributeReadOnlyConfig,
)
from tango.pyaml.controlsystem import (
    ConfigModel as ControlSystemConfig,
)
from tango.pyaml.controlsystem import (
    TangoControlSystem,
)

from pyaml.accelerator import Accelerator
from pyaml.accelerator import ConfigModel as AcceleratorConfigModel
from pyaml.arrays.magnet import ConfigModel as MagnetArrayConfigModel
from pyaml.arrays.magnet import Magnet
from pyaml.configuration import set_root_folder
from pyaml.configuration.csvcurve import ConfigModel as CSVCureveConfig
from pyaml.configuration.csvcurve import CSVCurve
from pyaml.lattice.simulator import ConfigModel as SimulatorConfigModel
from pyaml.lattice.simulator import Simulator
from pyaml.magnet.linear_model import (
    ConfigModel as LinearMagnetModelConfig,
)
from pyaml.magnet.linear_model import (
    LinearMagnetModel,
)
from pyaml.magnet.quadrupole import ConfigModel as QuadrupoleConfig
from pyaml.magnet.quadrupole import Quadrupole

# Get the directory of the current script
script_dir = os.path.dirname(__file__)

# Go up one level and then into 'data'
relative_path = os.path.join(script_dir, "..", "..", "tests")

# Normalize the path (resolves '..')
absolute_path = os.path.abspath(relative_path)

set_root_folder(absolute_path)

# Configuration

tangocs = ControlSystemConfig(name="live", tango_host="ebs-simu-3:10000")
control = TangoControlSystem(tangocs)

qfCurve = CSVCurve(CSVCureveConfig(file="config/sr/magnet_models/QF1_strength.csv"))
qdCurve = CSVCurve(CSVCureveConfig(file="config/sr/magnet_models/QD2_strength.csv"))

elemConfig = [
    {
        "name": "QF1A-C01",
        "attname": "srmag/vps-qf1/c01-a/current",
        "calibration_factor": 1.00504,
        "curve": qfCurve,
    },
    {
        "name": "QF1E-C01",
        "attname": "srmag/vps-qf1/c01-e/current",
        "calibration_factor": 0.998212,
        "curve": qfCurve,
    },
    {
        "name": "QD2A-C01",
        "attname": "srmag/vps-qd2/c01-a/current",
        "calibration_factor": 1.00504,
        "curve": qdCurve,
    },
    {
        "name": "QD2E-C01",
        "attname": "srmag/vps-qd2/c01-e/current",
        "calibration_factor": 1.003485189,
        "curve": qdCurve,
    },
]

devices = []
names = []
for cfg in elemConfig:
    qAtt = Attribute(AttributeConfig(attribute=cfg["attname"], unit="A"))
    qModel = LinearMagnetModel(
        LinearMagnetModelConfig(
            curve=cfg["curve"],
            calibration_factor=cfg["calibration_factor"],
            powerconverter=qAtt,
            unit="1/m",
        )
    )
    devices.append(Quadrupole(QuadrupoleConfig(name=cfg["name"], model=qModel)))
    names.append(cfg["name"])


simulator = Simulator(
    SimulatorConfigModel(name="design", lattice="config/sr/lattices/ebs.mat")
)

quads = Magnet(MagnetArrayConfigModel(name="quadsForTune", elements=names))

sr = Accelerator(
    AcceleratorConfigModel(
        machine="sr",
        facility="ESRF",
        energy=6e9,
        controls=[control],
        simulators=[simulator],
        devices=devices,
        arrays=[quads],
        data_folder="/tmp",
    )
)


# Usage exmaple

quadForTuneDesign = sr.design.get_magnets("quadsForTune")
quadForTuneLive = sr.live.get_magnets("quadsForTune")

# Compute tune response matrix for the 4 quads from simulator
sr.design.get_lattice().disable_6d()
tune = sr.design.get_lattice().get_tune()
tunemat = np.zeros((len(quadForTuneDesign), 2))
for idx, m in enumerate(quadForTuneDesign):
    str = m.strength.get()
    m.strength.set(str + 1e-4)
    dq = sr.design.get_lattice().get_tune() - tune
    tunemat[idx] = dq * 1e4
    m.strength.set(str)

# Compute correction matrix
correctionmat = np.linalg.pinv(tunemat.T)

# Correct tune on live

# Here ,as an example, instead of using the betatron tune monitor object,
# we directly connect to th CS
qxAtt, qyAtt = control.attach(
    [
        AttributeReadOnly(
            AttributeReadOnlyConfig(attribute="sys/ringsimulator/ebs/Tune_h", unit="")
        ),
        AttributeReadOnly(
            AttributeReadOnlyConfig(attribute="sys/ringsimulator/ebs/Tune_v", unit="")
        ),
    ]
)

print(f"Tune={qxAtt.readback()}, {qyAtt.readback()}")

strs = quadForTuneLive.strengths.get()
strs += np.matmul(correctionmat, [0.1, 0.05])  # Ask for correction [dqx,dqy]
quadForTuneLive.strengths.set(strs)
time.sleep(3)
print(f"Tune={qxAtt.readback()}, {qyAtt.readback()}")



from scipy.constants import speed_of_light

from ..common import abstract
from ..common.abstract import RWMapper
from ..common.element import Element, ElementConfigModel, __pyaml_repr__
from ..common.exception import PyAMLException
from ..configuration import Factory
from .hcorrector import HCorrector
from .magnet import Magnet, MagnetConfigModel
from .model import MagnetModel
from .octupole import Octupole
from .quadrupole import Quadrupole
from .sextupole import Sextupole
from .skewoctu import SkewOctu
from .skewquad import SkewQuad
from .skewsext import SkewSext
from .vcorrector import VCorrector

_fmap: dict = {
    "B0": HCorrector,
    "A0": VCorrector,
    "B1": Quadrupole,
    "A1": SkewQuad,
    "B2": Sextupole,
    "A2": SkewSext,
    "B3": Octupole,
    "A3": SkewOctu,
}

# Define the main class name for this module
PYAMLCLASS = "CombinedFunctionMagnet"


class ConfigModel(ElementConfigModel):
    mapping: list[list[str]]
    """Name mapping for multipoles
    (i.e. [[B0,C01A-H],[A0,C01A-H],[B2,C01A-S]])"""
    model: MagnetModel | None = None
    """Object in charge of converting magnet strenghts to currents"""


class CombinedFunctionMagnet(Element):
    """CombinedFunctionMagnet class"""

    def __init__(self, cfg: ConfigModel, peer=None):
        super().__init__(cfg.name)
        self._cfg = cfg
        self.model = cfg.model
        self.__virtuals: list[Magnet] = []
        self.__strengths: abstract.ReadWriteFloatArray = None
        self.__hardwares: abstract.ReadWriteFloatArray = None

        if peer is None:
            # Configuration part
            if self.model is not None and not hasattr(self.model._cfg, "multipoles"):
                raise PyAMLException(
                    f"{cfg.name} model: mutipoles"
                    "field required for combined function magnet"
                )

            idx = 0
            self.polynoms = []
            for _idx, m in enumerate(cfg.mapping):
                # Check mapping validity
                if len(m) != 2:
                    raise PyAMLException(
                        "Invalid CombinedFunctionMagnet mapping for {m}"
                    )
                if m[0] not in _fmap:
                    raise PyAMLException(
                        m[0] + " not implemented for combined function magnet"
                    )
                if m[0] not in self.model._cfg.multipoles:
                    raise PyAMLException(m[0] + " not found in underlying magnet model")
                self.polynoms.append(_fmap[m[0]].polynom)
                # Create the virtual magnet for the correspoding multipole
                vm = self.__create_virutal_manget(m[1], m[0])
                self.__virtuals.append(vm)
                # Register the virtual element in the factory to have
                # a coherent factory and improve error reporting
                Factory.register_element(vm)

        else:
            # Attach
            self._peer = peer

    def get_model_name(self) -> str:
        """
        Returns the model name of this magnet
        """
        return self._cfg.name

    def __create_virutal_manget(self, name: str, idx: int) -> Magnet:
        args = {"name": name, "model": self.model}
        mVirtual: Magnet = _fmap[idx](MagnetConfigModel(**args))
        mVirtual.set_model_name(self.get_name())
        return mVirtual

    def nb_multipole(self) -> int:
        return len(self._cfg.mapping)

    def attach(
        self,
        peer,
        strengths: abstract.ReadWriteFloatArray,
        hardwares: abstract.ReadWriteFloatArray,
    ) -> list[Magnet]:
        l = []
        # Attached the CombinedFunctionMagnet itself
        nCFM = CombinedFunctionMagnet(self._cfg, peer)
        nCFM.__strengths = strengths
        nCFM.__hardwares = hardwares
        l.append(nCFM)
        # Construct a single function magnet for each multipole
        # of this combined function magnet
        for idx, _m in enumerate(self._cfg.mapping):
            strength = RWMapper(strengths, idx)
            hardware = RWMapper(hardwares, idx) if self.model.has_hardware() else None
            l.append(self.__virtuals[idx].attach(peer, strength, hardware))
        return l

    @property
    def strengths(self) -> abstract.ReadWriteFloatScalar:
        """
        Gives access to the strengths of this combined
        function magnet in physics unit
        """
        self.check_peer()
        if self.__strengths is None:
            raise PyAMLException(
                f"{str(self)} has no model that supports physics units"
            )
        return self.__strengths

    @property
    def hardwares(self) -> abstract.ReadWriteFloatScalar:
        """
        Gives access to the strengths of this combined
        function magnet in hardware unit when possible
        """
        self.check_peer()
        if self.__hardwares is None:
            raise PyAMLException(
                f"{str(self)} has no model that supports hardware units"
            )
        return self.__hardwares

    def set_energy(self, E: float):
        if self.model is not None:
            self.model.set_magnet_rigidity(E / speed_of_light)

    def __repr__(self):
        return __pyaml_repr__(self)



from datetime import datetime
from enum import Enum, auto
from typing import Union

import numpy as np


class Quality(Enum):
    VALID = auto()
    INVALID = auto()
    ALARM = auto()
    CHANGING = auto()
    WARNING = auto()

    def __str__(self):
        return self.name


class Value:
    """
    Represents a numerical value with associated quality and timestamp.

    This class wraps a scalar or NumPy array and can be used in arithmetic
    operations and comparisons. It also retains metadata like quality and timestamp.

    Parameters
    ----------
    value : float, int or numpy.ndarray
        The numerical value(s) to wrap.
    quality : Quality, optional
        The quality of the value. Defaults to Quality.VALID.
    timestamp : datetime, optional
        Timestamp associated with the value. Defaults to current time.
    """

    def __init__(
        self,
        value: Union[float, int, np.ndarray],
        quality: Quality = Quality.VALID,
        timestamp: datetime = None,
    ):
        self.value = value
        self.quality = quality
        self.timestamp = timestamp or datetime.now()

    def __repr__(self):
        """
        Return a string representation of the Value object.

        Returns
        -------
        str
            Human-readable representation.
        """
        return (
            f"Value({self.value}, quality='{self.quality}',"
            + f"timestamp='{self.timestamp}')"
        )

    def __float__(self):
        """
        Convert to float.

        Returns
        -------
        float
            The value converted to float.
        """
        return float(self.value)

    def __int__(self):
        """
        Convert to int.

        Returns
        -------
        int
            The value converted to int.
        """
        return int(self.value)

    def __eq__(self, other):
        """
        Compare for equality.

        Parameters
        ----------
        other : Value, scalar or numpy.ndarray
            The object to compare against.

        Returns
        -------
        bool
            True if values are equal, False otherwise.
        """
        if isinstance(other, Value):
            return np.array_equal(self.value, other.value)
        if isinstance(self.value, np.ndarray):
            return np.array_equal(self.value, other)
        return self.value == other

    def __lt__(self, other):
        """
        Less than comparison.

        Parameters
        ----------
        other : Value or scalar

        Returns
        -------
        bool
        """
        other_val = other.value if isinstance(other, Value) else other
        return self.value < other_val

    def __le__(self, other):
        """
        Less than or equal comparison.

        Parameters
        ----------
        other : Value or scalar

        Returns
        -------
        bool
        """
        other_val = other.value if isinstance(other, Value) else other
        return self.value <= other_val

    def __gt__(self, other):
        """
        Greater than comparison.

        Parameters
        ----------
        other : Value or scalar

        Returns
        -------
        bool
        """
        other_val = other.value if isinstance(other, Value) else other
        return self.value > other_val

    def __ge__(self, other):
        """
        Greater than or equal comparison.

        Parameters
        ----------
        other : Value or scalar

        Returns
        -------
        bool
        """
        other_val = other.value if isinstance(other, Value) else other
        return self.value >= other_val

    def __add__(self, other):
        """
        Add another value.

        Parameters
        ----------
        other : Value or scalar or numpy.ndarray
            Value to add.

        Returns
        -------
        Result of addition.
        """
        other_val = other.value if isinstance(other, Value) else other
        return self.value + other_val

    def __radd__(self, other):
        """
        Add from the right-hand side.

        Parameters
        ----------
        other : scalar or numpy.ndarray
            Value to add.

        Returns
        -------
        Result of addition.
        """
        return other + self.value

    def __sub__(self, other):
        """
        Subtract another value.

        Parameters
        ----------
        other : Value or scalar or numpy.ndarray
            Value to subtract.

        Returns
        -------
        Result of subtraction.
        """
        other_val = other.value if isinstance(other, Value) else other
        return self.value - other_val

    def __rsub__(self, other):
        """
        Subtract from the right-hand side.

        Parameters
        ----------
        other : scalar or numpy.ndarray
            Value to subtract.

        Returns
        -------
        Result of subtraction.
        """
        return other - self.value

    def __mul__(self, other):
        """
        Multiply by another value.

        Parameters
        ----------
        other : Value or scalar or numpy.ndarray
            Value to multiply with.

        Returns
        -------
        Result of multiplication.
        """
        other_val = other.value if isinstance(other, Value) else other
        return self.value * other_val

    def __rmul__(self, other):
        """
        Multiply from the right-hand side.

        Parameters
        ----------
        other : scalar or numpy.ndarray
            Value to multiply with.

        Returns
        -------
        Result of multiplication.
        """
        return other * self.value

    def __truediv__(self, other):
        """
        Divide by another value.

        Parameters
        ----------
        other : Value or scalar or numpy.ndarray
            Divisor.

        Returns
        -------
        Result of division.
        """
        other_val = other.value if isinstance(other, Value) else other
        return self.value / other_val

    def __rtruediv__(self, other):
        """
        Divide from the right-hand side.

        Parameters
        ----------
        other : scalar or numpy.ndarray
            Dividend.

        Returns
        -------
        Result of division.
        """
        return other / self.value

    def __neg__(self):
        """
        Negate the value.

        Returns
        -------
        Result of negation.
        """
        return -self.value

    def is_good(self) -> bool:
        """
        Check if the value quality is good.

        Returns
        -------
        bool
            True if the quality is VALID or CHANGING.
        """
        return self.quality in (Quality.VALID, Quality.CHANGING)



import numpy as np
from numpy.typing import NDArray
from pydantic import BaseModel, ConfigDict

from pyaml.bpm.bpm_model import BPMModel

from ..common.element import __pyaml_repr__
from ..control.deviceaccess import DeviceAccess

# Define the main class name for this module
PYAMLCLASS = "BPMSimpleModel"


class ConfigModel(BaseModel):
    """
    Configuration model for BPM simple model

    Parameters
    ----------
    x_pos : DeviceAccess, optional
        Horizontal position device
    y_pos : DeviceAccess, optional
        Vertical position device
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    x_pos: DeviceAccess | None
    y_pos: DeviceAccess | None


class BPMSimpleModel(BPMModel):
    """
    Concrete implementation of BPMModel that simulates a BPM with tilt and
    offset values.
    """

    def __init__(self, cfg: ConfigModel):
        self._cfg = cfg
        self.__x_pos = cfg.x_pos
        self.__y_pos = cfg.y_pos

    def get_pos_devices(self) -> list[DeviceAccess | None]:
        """
        Get device handles used for position reading

        Returns
        -------
        list[DeviceAccess]
            Array of DeviceAcess
        """
        return [self.__x_pos, self.__y_pos]

    def get_tilt_device(self) -> DeviceAccess | None:
        """
        Get device handle used for tilt access

        Returns
        -------
        list[DeviceAccess]
            Array of DeviceAcess
        """
        return None

    def get_offset_devices(self) -> list[DeviceAccess | None]:
        """
        Get device handles used for offset access

        Returns
        -------
        list[DeviceAccess]
            Array of DeviceAcess
        """
        return [None, None]

    def __repr__(self):
        return __pyaml_repr__(self)



from pathlib import Path

from pyaml.accelerator import Accelerator
from pyaml.tuning_tools.orbit_response_matrix import ConfigModel as ORM_ConfigModel
from pyaml.tuning_tools.orbit_response_matrix import OrbitResponseMatrix

parent_folder = Path(__file__).parent
config_path = parent_folder.parent.parent.joinpath(
    "tests", "config", "EBSOrbit.yaml"
).resolve()
sr = Accelerator.load(config_path)
element_holder = sr.design

orm = OrbitResponseMatrix(
    cfg=ORM_ConfigModel(
        bpm_array_name="BPM",
        hcorr_array_name="HCorr",
        vcorr_array_name="VCorr",
        corrector_delta=1e-6,
    ),
    element_holder=element_holder,
)

hcorr = element_holder.get_magnets("HCorr")
vcorr = element_holder.get_magnets("VCorr")
corrector_names = (
    hcorr["SJ2A*"].names()
    + hcorr["SF2A*"].names()
    + hcorr["SI2A*"].names()
    + vcorr["SJ2A*"].names()
    + vcorr["SF2A*"].names()
    + vcorr["SI2A*"].names()
)

orm.measure(corrector_names=corrector_names)
orm.save(parent_folder / Path("reduced_orm.json"))

ormdata = orm.get()



import at
from pydantic import ConfigDict

from pyaml.common.element import Element
from pyaml.lattice.lattice_elements_linker import (
    LatticeElementsLinker,
    LinkerConfigModel,
    LinkerIdentifier,
)

PYAMLCLASS = "PyAtAttributeElementsLinker"


class ConfigModel(LinkerConfigModel):
    """Base configuration model for linker definitions.

    This class defines the configuration structure used to instantiate
    a specific linking strategy. Each concrete implementation of a
    `LatticeElementsLinker` may define its own subclass extending this model
    to include additional configuration parameters.

    Attributes
    ----------
    model_config : ConfigDict
        Pydantic configuration allowing arbitrary field types and forbidding
        unexpected extra keys.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")
    attribute_name: str


class PyAtAttributeIdentifier(LinkerIdentifier):
    """Abstract base class for identifiers used to match PyAML and PyAT elements.

    The identifier acts as an intermediate representation between the PyAML
    configuration and the PyAT lattice. Its exact structure depends on the
    linking strategy (e.g., family name, element index, or user-defined tag).

    Subclasses should define the fields and logic necessary to represent
    a unique reference to one or more PyAT elements.
    """

    def __init__(self, attribute_name: str, identifier):
        self.attribute_name = attribute_name
        self.identifier = identifier

    def __repr__(self):
        return f"{self.attribute_name}={self.identifier}"


class PyAtAttributeElementsLinker(LatticeElementsLinker):
    """Abstract base class defining the interface for PyAT–PyAML element linking.

    Implementations of this class define how PyAML elements are matched
    to PyAT elements based on a given linking strategy (e.g., by family name,
    by index, or by a custom attribute).

    Parameters
    ----------
    config_model : ConfigModel
        The configuration model for the linking strategy.
    """

    def __init__(self, config_model: ConfigModel):
        super().__init__(config_model)

    def get_element_identifier(self, element: Element) -> LinkerIdentifier:
        return PyAtAttributeIdentifier(
            self.linker_config_model.attribute_name, element.get_name()
        )

    def _test_at_element(
        self, identifier: PyAtAttributeIdentifier, element: at.Element
    ) -> bool:
        attr_value = getattr(element, identifier.attribute_name, None)
        return attr_value == identifier.identifier



import numpy as np
from pydantic import BaseModel, ConfigDict

try:
    from typing import Self  # Python 3.11+
except ImportError:
    from typing_extensions import Self  # Python 3.10 and earlier

from .. import PyAMLException
from ..common import abstract
from ..common.element import Element, ElementConfigModel
from ..control.deviceaccess import DeviceAccess
from .rf_transmitter import RFTransmitter

# Define the main class name for this module
PYAMLCLASS = "RFPlant"


class ConfigModel(ElementConfigModel):
    masterclock: DeviceAccess | None = None
    """Device to apply main RF frequency"""
    transmitters: list[RFTransmitter] | None = None
    """List of RF trasnmitters"""


class RFPlant(Element):
    """
    Main RF object
    """

    def __init__(self, cfg: ConfigModel):
        super().__init__(cfg.name)
        self._cfg = cfg
        self.__frequency = None
        self.__voltage = None

    @property
    def frequency(self) -> abstract.ReadWriteFloatScalar:
        if self.__frequency is None:
            raise PyAMLException(f"{str(self)} has no masterclock device defined")
        return self.__frequency

    @property
    def voltage(self) -> abstract.ReadWriteFloatScalar:
        if self.__voltage is None:
            raise PyAMLException(f"{str(self)} has no trasmitter device defined")
        return self.__voltage

    def attach(
        self,
        peer,
        frequency: abstract.ReadWriteFloatScalar,
        voltage: abstract.ReadWriteFloatScalar,
    ) -> Self:
        # Attach frequency attribute and returns a new reference
        obj = self.__class__(self._cfg)
        obj.__frequency = frequency
        obj.__voltage = voltage
        obj._peer = peer
        return obj


class RWTotalVoltage(abstract.ReadWriteFloatScalar):
    def __init__(self, transmitters: list[RFTransmitter]):
        """
        Construct a RWTotalVoltage setter

        Parameters
        ----------
        transmitters : list[RFTransmitter]
            List of attached transmitters
        """
        self.__trans = transmitters

    def get(self) -> float:
        sum = 0
        # Count only fundamental harmonic
        for t in self.__trans:
            if t._cfg.harmonic == 1.0:
                sum += t.voltage.get()
        return sum

    def set(self, value: float):
        # Assume that sum of transmitter (fundamental harmonic) distribution is 1
        for t in self.__trans:
            if t._cfg.harmonic == 1.0:
                v = value * t._cfg.distribution
                t.voltage.set(v)

    def set_and_wait(self, value: float):
        raise NotImplementedError("Not implemented yet.")

    def unit(self) -> str:
        return self.__trans[0]._cfg.phase.unit()



from pathlib import Path

import numpy as np
from pydantic import BaseModel, ConfigDict

from ..common.exception import PyAMLException
from ..configuration import get_root_folder
from .curve import Curve

# Define the main class name for this module
PYAMLCLASS = "InlineCurve"


class ConfigModel(BaseModel):
    """
    Configuration model for inline curve

    Parameters
    ----------
    mat : list[list[float]]
        Curve data (n rows, 2 columns)
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    mat: list[list[float]]


class InlineCurve(Curve):
    """
    Class for load CSV (x,y) curve
    """

    def __init__(self, cfg: ConfigModel):
        self._cfg = cfg
        # Load the curve
        self._curve = np.array(self._cfg.mat)

        _s = np.shape(self._curve)
        if len(_s) != 2 or _s[1] != 2:
            raise PyAMLException(
                f"InlineCurve(mat='{cfg.mat}',dtype=float): wrong shape (2,2) expected but got {str(_s)}"
            )

    def get_curve(self) -> np.array:
        return self._curve

    def __repr__(self):
        return repr(self._cfg).replace("ConfigModel", self.__class__.__name__)



from ..common.element_holder import ElementHolder
from .array import ArrayConfig, ArrayConfigModel

# Define the main class name for this module
PYAMLCLASS = "CombinedFunctionMagnet"


class ConfigModel(ArrayConfigModel): ...


class CombinedFunctionMagnet(ArrayConfig):
    """
    Combined function magnet array confirguration

    Example
    -------

    A magnet array configuration can also be created by code using
    the following example::

        from pyaml.arrays.cfm_magnet import CombinedFunctionMagnet
        from pyaml.arrays.cfm_magnet import ConfigModel as CFMagnetConfigModel
        magArray = CombinedFunctionMagnet(
                     CFMagnetConfigModel(name="myCFM", elements=["mag1","mag2"])
                   )
    """

    def __init__(self, cfg: ArrayConfigModel):
        super().__init__(cfg)

    def fill_array(self, holder: ElementHolder):
        holder.fill_cfm_magnet_array(self._cfg.name, self._cfg.elements)



from ..common.element_holder import ElementHolder
from .array import ArrayConfig, ArrayConfigModel

# Define the main class name for this module
PYAMLCLASS = "SerializedMagnets"


class ConfigModel(ArrayConfigModel): ...


class SerializedMagnets(ArrayConfig):
    """
    Serialized magnets array configuration

    Example
    -------

    A magnet array configuration can also be created by code using
    the following example::

        from pyaml.arrays.serialized_magnet import SerializedMagnets
        from pyaml.arrays.serialized_magnet import ConfigModel as SerializedMagnetConfigModel
        magArray = SerializedMagnets(
                     SerializedMagnetConfigModel(name="mySerializedMagnets", elements=["mag1","mag2"])
                   )
    """

    def __init__(self, cfg: ArrayConfigModel):
        super().__init__(cfg)

    def fill_array(self, holder: ElementHolder):
        holder.fill_serialized_magnet_array(self._cfg.name, self._cfg.elements)



from typing import TYPE_CHECKING, Optional, Tuple

import numpy as np

if TYPE_CHECKING:
    from ..common.element_holder import ElementHolder
from ..common.exception import PyAMLException


class pySCInterface:
    def __init__(
        self,
        element_holder: "ElementHolder",
        bpm_array_name: str,
        rf_plant_name: Optional[str] = None,
    ):
        self.element_holder = element_holder

        self.bpm_array = element_holder.get_bpms(bpm_array_name)

        self.rf_plant_name = rf_plant_name
        if rf_plant_name is not None:
            self.rf_plant = element_holder.get_rf_plant(self.rf_plant_name)
        else:
            self.rf_plant = None

    def get_orbit(self) -> Tuple[np.array, np.array]:
        # we should wait here somehow according to polling rate
        positions = self.bpm_array.positions.get()
        return positions[:, 0], positions[:, 1]

    def get(self, name: str) -> float:
        magnet = self.element_holder.get_magnet(name=name)
        return magnet.strength.get()

    def set(self, name: str, value: float) -> None:
        magnet = self.element_holder.get_magnet(name=name)
        magnet.strength.set(value=value)  # ideally set_and_wait but not implemented
        return

    def get_rf_main_frequency(self) -> float:
        if self.rf_plant is None:
            raise PyAMLException("RF plant name was not provided.")
        return self.rf_plant.frequency.get()

    def set_rf_main_frequency(self, value: float) -> None:
        if self.rf_plant is None:
            raise PyAMLException("RF plant name was not provided.")
        self.rf_plant.frequency.set(value)
        return



from abc import ABCMeta, abstractmethod

import numpy as np


class Curve(metaclass=ABCMeta):
    """
    Abstract class providing access to a curve
    """

    @abstractmethod
    def get_curve(self) -> np.array:
        """
        Returns the curve (n rows,2 columns).
        Curve is expected to be monotonic (non-decreasing or non-increasing).
        """
        pass

    @classmethod
    def inverse(cls, curve: np.array) -> np.array:
        """
        Returns the inverse curve.
        Basically swap x and y and sort y in ascending order.

        Parameters
        ----------
        curve : np.array
            Curve to be inverted
        """
        __curve = curve
        __sortedCurve = __curve[__curve[:, 1].argsort()]
        return __sortedCurve[:, [1, 0]]



import os

import numpy as np

from pyaml.accelerator import Accelerator
from pyaml.common.constants import ACTION_RESTORE
from pyaml.magnet.magnet import Magnet

# Get the directory of the current script
script_dir = os.path.dirname(__file__)

# Go up one level and then into 'data'
relative_path = os.path.join(script_dir, "..", "..", "tests", "config", "EBSTune.yaml")

# Normalize the path (resolves '..')
absolute_path = os.path.abspath(relative_path)

sr: Accelerator = Accelerator.load(absolute_path)
sr.design.get_lattice().disable_6d()


# Callback exectued after each magnet strenght setting
# during the tune response matrix measurement
def tune_callback(step: int, action: int, m: Magnet, dtune: np.array):
    if action == ACTION_RESTORE:
        # On action restore, the delta tune is passed as argument
        print(f"Tune response: #{step} {m.get_name()} {dtune}")
    return True


# Compute tune response matrix
tune_adjust_design = sr.design.get_tune_tuning("TUNE")
tune_adjust_design.response.measure(callback=tune_callback)
tune_adjust_design.response.save_json("tunemat.json")

# Correct tune on live
tune_adjust_live = sr.live.get_tune_tuning("TUNE")
tune_adjust_live.response.load_json("tunemat.json")
print(tune_adjust_live.readback())
tune_adjust_live.set([0.17, 0.32], iter=2, wait_time=10)
print(tune_adjust_live.readback())



import numpy as np
from pydantic import BaseModel, ConfigDict

from ..common.element import __pyaml_repr__
from ..configuration.curve import Curve
from ..control.deviceaccess import DeviceAccess
from .model import MagnetModel

# Define the main class name for this module
PYAMLCLASS = "LinearMagnetModel"


class ConfigModel(BaseModel):
    """
    Linear magnet model.

    Parameters
    ----------
    curve : Curve or None, optional
        Curve object used for interpolation. By default,
        identity curve is used.
    powerconverter : DeviceAccess or None, optional
        Power converter device to apply currrent
    calibration_factor : float, optional
        Correction factor applied to the curve. Default: 1.0
    calibration_offset : float, optional
        Correction offset applied to the curve. Default: 0.0
    crosstalk : float, optional
        Crosstalk factor. Default: 1.0
    unit : str
        Unit of the strength (i.e. 1/m or m-1)

    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    curve: Curve | None = None
    powerconverter: DeviceAccess | None
    calibration_factor: float = 1.0
    calibration_offset: float = 0.0
    crosstalk: float = 1.0
    unit: str


class LinearMagnetModel(MagnetModel):
    """
    Class that handle manget current/strength conversion using
    linear interpolation for a single function magnet
    """

    def __init__(self, cfg: ConfigModel):
        self._cfg = cfg
        if self._cfg.curve:
            self.__curve = cfg.curve.get_curve()
            self.__curve[:, 1] = (
                self.__curve[:, 1] * cfg.calibration_factor * cfg.crosstalk
                + cfg.calibration_offset
            )
            self.__rcurve = Curve.inverse(self.__curve)
        else:
            self.__curve = None
            self.__rcurve = None
            self.__g = cfg.calibration_factor * cfg.crosstalk
            self.__o = cfg.calibration_offset
        self.__strength_unit = cfg.unit
        self.__hardware_unit = cfg.powerconverter.unit()
        self.__brho = np.nan
        self.__ps = cfg.powerconverter

    def compute_hardware_values(self, strengths: np.array) -> np.array:
        if self.__rcurve is not None:
            _current = np.interp(
                strengths[0] * self.__brho, self.__rcurve[:, 0], self.__rcurve[:, 1]
            )
        else:
            _current = (strengths[0] * self.__brho) / self.__g + self.__o
        return np.array([_current])

    def compute_strengths(self, currents: np.array) -> np.array:
        if self.__curve is not None:
            _strength = (
                np.interp(currents[0], self.__curve[:, 0], self.__curve[:, 1])
                / self.__brho
            )
        else:
            _strength = ((currents[0] - self.__o) * self.__g) / self.__brho
        return np.array([_strength])

    def get_strength_units(self) -> list[str]:
        return [self.__strength_unit] if self.__strength_unit is not None else [""]

    def get_hardware_units(self) -> list[str]:
        return [self.__hardware_unit] if self.__hardware_unit is not None else [""]

    def get_devices(self) -> list[DeviceAccess]:
        return [self.__ps]

    def set_magnet_rigidity(self, brho: np.double):
        self.__brho = brho

    def __repr__(self):
        return __pyaml_repr__(self)



from ..common.abstract import ReadFloatArray
from ..common.element import Element, ElementConfigModel
from ..control.deviceaccess import DeviceAccess

try:
    from typing import Self  # Python 3.11+
except ImportError:
    from typing_extensions import Self  # Python 3.10 and earlier
from pydantic import ConfigDict

PYAMLCLASS = "BetatronTuneMonitor"


class ConfigModel(ElementConfigModel):
    """
    Configuration model for BetatronTuneMonitor

    Parameters
    ----------
    tune_h : DeviceAccess, optional
        Horizontal betatron tune device
    tune_v : DeviceAccess, optional
        Vertical betatron tune device
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    tune_h: DeviceAccess | None
    tune_v: DeviceAccess | None


class BetatronTuneMonitor(Element):
    """
    Class providing access to a betatron tune monitor
    of a physical or simulated lattice.
    The monitor provides horizontal and vertical betatron tune measurements.
    """

    def __init__(self, cfg: ConfigModel):
        """
        Construct a BetatronTuneMonitor

        Parameters
        ----------
        cfg : ConfigModel
            Configuration for the BetatronTuneMonitor, including
            device access for horizontal and vertical tunes.
        """

        super().__init__(cfg.name)
        self._cfg = cfg
        self.__tune = None

    @property
    def tune(self) -> ReadFloatArray:
        self.check_peer()
        return self.__tune

    def attach(self, peer, betatron_tune: ReadFloatArray) -> Self:
        obj = self.__class__(self._cfg)
        obj.__tune = betatron_tune
        obj._peer = peer
        return obj



import numpy as np
from pydantic import BaseModel, ConfigDict

from .. import PyAMLException
from ..common.element import __pyaml_repr__
from ..control.deviceaccess import DeviceAccess
from .model import MagnetModel

# Define the main class name for this module
PYAMLCLASS = "IdentityCFMagnetModel"


class ConfigModel(BaseModel):
    """
    Configuration model for identity combined function magnet model

    Parameters
    ----------
    multipoles : list[str]
        List of supported functions: A0, B0, A1, B1, etc (i.e. [B0, A1, B2])
    powerconverters : list[DeviceAccess], optional
        Power converter devices to apply current
    physics : list[DeviceAccess], optional
        Magnet devices to apply strength
    units : list[str]
        List of strength units (i.e. ['rad', 'm-1', 'm-2'])
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    multipoles: list[str]
    powerconverters: list[DeviceAccess | None] | None = None
    physics: list[DeviceAccess | None] | None = None
    units: list[str]


class IdentityCFMagnetModel(MagnetModel):
    """
    Class that map values to underlying devices without conversion
    """

    def __init__(self, cfg: ConfigModel):
        self._cfg = cfg

        # Check config
        self.__nbFunction: int = len(cfg.multipoles)

        if cfg.physics is None and cfg.powerconverters is None:
            raise PyAMLException(
                "Invalid IdentityCFMagnetModel configuration,"
                "physics or powerconverters device required"
            )
        if cfg.physics is not None and cfg.powerconverters is not None:
            raise PyAMLException(
                "Invalid IdentityCFMagnetModel configuration,"
                "physics or powerconverters device required but not both"
            )
        if cfg.physics:
            self.__devices = cfg.physics
        else:
            self.__devices = cfg.powerconverters

        self.__nbDev: int = len(self.__devices)

        self.__check_len(cfg.units, "units", self.__nbFunction)

    def __check_len(self, obj, name, expected_len):
        lgth = len(obj)
        if lgth != expected_len:
            raise PyAMLException(
                f"{name} does not have the expected "
                f"number of items ({expected_len} items expected but got {lgth})"
            )

    def compute_hardware_values(self, strengths: np.array) -> np.array:
        return strengths

    def compute_strengths(self, currents: np.array) -> np.array:
        return currents

    def get_strength_units(self) -> list[str]:
        return self._cfg.units

    def get_hardware_units(self) -> list[str]:
        return self._cfg.units

    def get_devices(self) -> list[DeviceAccess | None]:
        return self.__devices

    def set_magnet_rigidity(self, brho: np.double):
        pass

    def has_physics(self) -> bool:
        return self._cfg.physics is not None

    def has_hardware(self) -> bool:
        return self._cfg.powerconverters is not None

    def __repr__(self):
        return __pyaml_repr__(self)



from pathlib import Path

import numpy as np
from pydantic import BaseModel, ConfigDict

from ..common.exception import PyAMLException
from ..configuration.fileloader import get_path
from .matrix import Matrix

# Define the main class name for this module
PYAMLCLASS = "CSVMatrix"


class ConfigModel(BaseModel):
    """
    Configuration model for CSV matrix

    Parameters
    ----------
    file : str
        CSV file that contains the matrix
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    file: str


class CSVMatrix(Matrix):
    """
    Class for loading CSV matrix
    """

    def __init__(self, cfg: ConfigModel):
        self._cfg = cfg
        # Load CSV matrix
        path = get_path(cfg.file)
        try:
            self._mat = np.genfromtxt(path, delimiter=",", dtype=float, loose=False)
        except ValueError as e:
            raise PyAMLException(
                f"CSVMatrix(file='{cfg.file}',dtype=float): {str(e)}"
            ) from None

    def get_matrix(self) -> np.array:
        return self._mat

    def __repr__(self):
        return repr(self._cfg).replace("ConfigModel", self.__class__.__name__)



# PyAML config file loader
import collections.abc
import io
import json
import logging
import os
from pathlib import Path
from typing import TYPE_CHECKING, Union

import yaml
from yaml import CLoader
from yaml.constructor import ConstructorError
from yaml.loader import SafeLoader

from pyaml.configuration.factory import Factory

from .. import PyAMLException

if TYPE_CHECKING:
    from pyaml.accelerator import Accelerator


logger = logging.getLogger(__name__)

accepted_suffixes = [".yaml", ".yml", ".json"]


ROOT = {"path": Path.cwd().resolve()}


def set_root_folder(path: Union[str, Path]):
    """
    Set the root path for configuration files.
    """
    ROOT["path"] = Path(path)


def get_root_folder() -> Path:
    """
    Get the root path for configuration files.
    """
    return ROOT["path"]


def get_path(p: Path) -> Path:
    """
    Return unchanged input path if it is an absolute path,
    path relative to root folder otherwise.
    """
    if os.path.isabs(p):
        return p
    else:
        root = get_root_folder()
        return root / p


class PyAMLConfigCyclingException(PyAMLException):
    def __init__(self, error_filename: str, path_stack: list[Path]):
        self.error_filename = error_filename
        parent_file_stack = [parent_path.name for parent_path in path_stack]
        super().__init__(
            f"Circular file inclusion of {error_filename}. "
            f"File list before reaching it: {parent_file_stack}"
        )

    pass


def load(
    filename: str, paths_stack: list = None, use_fast_loader: bool = False
) -> Union[dict, list]:
    """Load recursively a configuration setup"""
    if filename.endswith(".yaml") or filename.endswith(".yml"):
        l = YAMLLoader(filename, paths_stack, use_fast_loader)
    elif filename.endswith(".json"):
        l = JSONLoader(filename, paths_stack, use_fast_loader)
    else:
        raise PyAMLException(
            f"{filename} File format not supported (only .yaml .yml or .json)"
        )
    return l.load()


# Expand condition
def hasToExpand(value):
    return isinstance(value, str) and any(
        value.endswith(suffix) for suffix in accepted_suffixes
    )


# Loader base class (nested files expansion)
class Loader:
    def __init__(self, filename: str, parent_path_stack: list[Path]):
        self.path: Path = get_path(filename)
        self.files_stack: list[Path] = []
        if parent_path_stack:
            if any(
                self.path.samefile(parent_path) for parent_path in parent_path_stack
            ):
                raise PyAMLConfigCyclingException(filename, parent_path_stack)
            self.files_stack.extend(parent_path_stack)
        self.files_stack.append(self.path)

    # Recursively expand a dict
    def expand_dict(self, d: dict):
        for key, value in d.items():
            try:
                if hasToExpand(value):
                    d[key] = load(value, self.files_stack, self.use_fast_loader)
                else:
                    self.expand(value)
            except PyAMLConfigCyclingException as pyaml_ex:
                location = d.pop("__location__", None)
                field_locations = d.pop("__fieldlocations__", None)
                location_str = ""
                if location:
                    file, line, col = location
                    if field_locations and key in field_locations:
                        location = field_locations[key]
                        file, line, col = location
                    location_str = f" in {file} at line {line}, column {col}"
                raise PyAMLException(
                    "Circular file inclusion "
                    f"of {pyaml_ex.error_filename}{location_str}"
                ) from pyaml_ex

    # Recursively expand a list
    def expand_list(self, l: list):
        for idx, value in enumerate(l):
            if hasToExpand(value):
                l[idx] = load(value, self.files_stack)
            else:
                self.expand(value)

    # Recursively expand an object
    def expand(self, obj: Union[dict, list]):
        if isinstance(obj, dict):
            self.expand_dict(obj)
        elif isinstance(obj, list):
            self.expand_list(obj)
        return obj

    # Load a file
    def load(self) -> Union[dict, list]:
        raise PyAMLException(str(self.path) + ": load() method not implemented")


class SafeLineLoader(SafeLoader):
    def __init__(self, stream):
        super().__init__(stream)
        self.filename = stream.name if isinstance(stream, io.TextIOWrapper) else ""

    def construct_mapping(self, node, deep=False):
        mapping = {}
        field_mapping = {}

        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=deep)
            if not isinstance(key, collections.abc.Hashable):
                raise ConstructorError(
                    "while constructing a mapping",
                    node.start_mark,
                    "found unhashable key",
                    key_node.start_mark,
                )
            value = self.construct_object(value_node, deep=deep)
            mapping[key] = value
            field_mapping[key] = (
                self.filename,
                value_node.start_mark.line + 1,
                value_node.start_mark.column + 1,
            )

        # Add location information inside the dict
        mapping["__location__"] = (
            self.filename,
            node.start_mark.line + 1,
            node.start_mark.column + 1,
        )
        mapping["__fieldlocations__"] = field_mapping
        return mapping


# YAML loader
class YAMLLoader(Loader):
    def __init__(self, filename: str, parent_paths_stack: list, use_fast_loader: bool):
        super().__init__(filename, parent_paths_stack)
        self._loader = SafeLineLoader if not use_fast_loader else CLoader
        self.use_fast_loader = use_fast_loader

    def load(self) -> Union[dict, list]:
        logger.log(logging.DEBUG, f"Loading YAML file '{self.path}'")
        with open(self.path) as file:
            try:
                return self.expand(yaml.load(file, Loader=self._loader))
            except yaml.YAMLError as e:
                raise PyAMLException(str(self.path) + ": " + str(e)) from e


# JSON loader
class JSONLoader(Loader):
    def __init__(self, filename: str, parent_paths_stack: list, use_fast_loader: bool):
        super().__init__(filename, parent_paths_stack)
        self.use_fast_loader = False

    def load(self) -> Union[dict, list]:
        logger.log(logging.DEBUG, f"Loading JSON file '{self.path}'")
        with open(self.path) as file:
            try:
                return self.expand(json.load(file))
            except json.JSONDecodeError as e:
                raise PyAMLException(str(self.path) + ": " + str(e)) from e
